/* Copyright (c) 1997 Andrew W. Appel.  Licensed software: see LICENSE file */
package Semant;

import Symbol.Symbol;
import Symbol.Table;
import Temp.Label;
import Temp.Temp;
import Translate.Level;
import Translate.BreakScope;
import ErrorMsg.ErrorMsg;
import Translate.Exp;
import Translate.ExpList;
import Types.Type;

class ExpTy {
   Exp exp;
   Type ty;
   ExpTy(Exp e, Type t) {exp=e; ty=t;}
}

public class Semant {

 Env env;
 Translate.Translate translate;
 Level level;
 BreakScope breakScope;

 public Semant(Translate.Translate t, ErrorMsg err) {
    this(new Env(t.outermost,err), t, t.outermost, (BreakScope)null);
 }

 Semant(Env e, Translate.Translate t, Level l, BreakScope b) {
    env=e;
    translate=t;
    level=l;
    breakScope=b;
 }

 void error(int pos, String msg) {env.errorMsg.error(pos,msg);}

 final static Types.INT INT = new Types.INT();
 final static Types.VOID VOID = new Types.VOID();
 ExpTy bogusExp() {return new ExpTy(translate.intExp(0), INT);}

 Types.Type tylook(Symbol id, int pos) {
     Types.Type ty = (Types.Type)(env.tenv.get(id));
     if (ty==null) {error(pos, "undefined type");
		    return VOID;}
     else return ty;
 } 

 void checkDups(Object p) {
 }

 Exp coerce(ExpTy a, Type t, int pos) {
      if (a.ty.coerceTo(t)) {}
          else error(pos, "type mismatch");
      return a.exp;
 }
    
 Exp checkInt(ExpTy a, int pos) {return coerce(a, INT, pos);}

 Type transTy(Absyn.NameTy t) {
 	return tylook(t.name,t.pos);
 }

 Types.RECORD transTypeFields(Absyn.FieldList f) {
     if (f==null) return null;
     else return new Types.RECORD(f.name, 
		              tylook(f.typ,f.pos), 
		              transTypeFields(f.tail));
 }


 Type transTy(Absyn.RecordTy t) {
     return transTypeFields(t.fields);
 }

 Type transTy(Absyn.ArrayTy t) {
     return new Types.ARRAY(tylook(t.typ, t.pos));
 }

 Type transTy(Absyn.Ty t) {
    if (t instanceof Absyn.ArrayTy) return transTy((Absyn.ArrayTy)t);
    if (t instanceof Absyn.NameTy) return transTy((Absyn.NameTy)t);
    if (t instanceof Absyn.RecordTy) return transTy((Absyn.RecordTy)t);
    throw new Error("transTy");
 }

  ExpTy transVar(Absyn.SimpleVar v) {
   Entry x = (Entry)env.venv.get(v.name);
   if (x instanceof VarEntry) {
      VarEntry ent = (VarEntry)x;
      return new ExpTy(translate.simpleVar(ent.access,level,v.name),ent.ty);
   }
   else {
     error(v.pos, "undefined variable");
     return bogusExp();
   }
 }

 ExpTy transVar(Absyn.FieldVar v) {
    ExpTy r = transVar(v.var);
    Types.Type rty = r.ty.actual();
    if (rty instanceof Types.RECORD) {
       int offset = 0;
       Types.RECORD t = (Types.RECORD)rty;
       for (; t!=null; offset++, t=t.tail)
          if (v.field == t.fieldName)
            return new ExpTy(translate.fieldVar(r.exp,offset,
						v.field), 
			     t.fieldType);
       error(v.pos, "field not present in record");
       return bogusExp();
    } else {
       error(v.pos, "selecting field from non-record");
       return bogusExp();
    }
 }

 ExpTy transVar(Absyn.SubscriptVar v) {
    ExpTy r = transVar(v.var);
    if (r.ty instanceof Types.ARRAY) {
        return new ExpTy(translate.subscriptVar(r.exp, 
                              checkInt(transExp(v.index), v.pos)),
	             ((Types.ARRAY)r.ty).element);
    } else {
        error(v.pos, "subscript of non-array");
        return bogusExp();
    }
 }

 ExpTy transVar(Absyn.Var v) {
   if (v instanceof Absyn.SimpleVar) return transVar((Absyn.SimpleVar)v);
   if (v instanceof Absyn.FieldVar) return transVar((Absyn.FieldVar)v);
   if (v instanceof Absyn.SubscriptVar) return transVar((Absyn.SubscriptVar)v);
   throw new Error("Semant.transVar");
 }

 ExpTy transExp(Absyn.VarExp e) {
    ExpTy a = transVar(e.var);
    return new ExpTy(translate.varExp(a.exp), a.ty);
 }

 ExpTy transExp(Absyn.IntExp e) {
    return new ExpTy(translate.intExp(e.value), INT);
 }

 ExpTy transExp(Absyn.StringExp e) {
    return new ExpTy(translate.stringExp(e.value), new Types.STRING());
 }

 ExpList zipArgs(Types.RECORD formals, Absyn.ExpList actuals) {
   if (formals==null) {
      if (actuals==null) 
         return null;
      else {
         error(actuals.head.pos,"too many actual parameters");
         return null;
      }
   } else
      if (actuals==null) {
         error(actuals.head.pos,"too few actual parameters");
         return null;
      } else
        return new ExpList(coerce(transExp(actuals.head),
	  			        formals.fieldType,
				        actuals.head.pos),
			         zipArgs(formals.tail, actuals.tail));
}
   
				   

 ExpTy transExp(Absyn.CallExp e) {
   Entry x = (Entry)env.venv.get(e.func);
   if (x instanceof FunEntry) {
      FunEntry ent = (FunEntry)x;
      return new ExpTy(translate.appExp(ent.label,level,ent.level,
				    zipArgs(ent.formals,e.args)),
		   ent.result);
   }
   else {
     error(e.pos, "Attempt to call non-function");
     return bogusExp();
   }
 }

 Exp notExp(Exp e) {
    return translate.ifThenElse(e,translate.intExp(0),translate.intExp(1));
 }

 ExpTy transExp(Absyn.OpExp e) {
   ExpTy left = transExp(e.left);
   ExpTy right = transExp(e.right);
   if (  e.oper == Absyn.OpExp.EQ || e.oper == Absyn.OpExp.NE)  {
     if (! (left.ty.coerceTo(right.ty) 
            || right.ty.coerceTo(left.ty)))
        error(e.pos, "type mismatch");
     if (left.ty.coerceTo(new Types.STRING())) {
       Exp eql = translate.stringEqual(left.exp,right.exp);
       if (e.oper == Absyn.OpExp.NE)
            return new ExpTy(notExp(eql), INT);
       else return new ExpTy(eql, INT);
     } else
       return new ExpTy(translate.opExp(e.oper,left.exp,right.exp),INT);
   } else 
     return new ExpTy(translate.opExp(e.oper, checkInt(left,e.pos),
				  checkInt(right, e.pos)),
	          INT);

 }

 ExpList zipFields(Types.RECORD formals, Absyn.FieldExpList actuals,
			     int pos) {
   if (formals==null) {
      if (actuals==null) 
         return null;
      else {
         error(actuals.pos,"too many fields");
         return null;
      }
   } else {
      if (actuals==null) {
         error(pos,"too few fields");
         return null;
      } else {
        if (actuals.name != formals.fieldName)
            error(actuals.pos, "field name mismatch");
        return new ExpList(coerce(transExp(actuals.init),
	  			        formals.fieldType,
				        actuals.pos),
			         zipFields(formals.tail, actuals.tail,
					   actuals.pos));
      }
   }
}

 ExpTy transExp(Absyn.RecordExp e) {
   Types.Type x = tylook(e.typ, e.pos).actual();
   if (x instanceof Types.RECORD) {
      Types.RECORD t = (Types.RECORD)x;
      return new ExpTy(translate.recordExp(zipFields(t, e.fields, e.pos)), t);
   } else {
      error(e.pos, "record-ID not a record type"); 
      return bogusExp();
   }
 } 
    
 ExpTy transExpSeq(Absyn.ExpList e) {
     if (e==null) return new ExpTy(translate.intExp(0), VOID);
     else if (e.tail==null) return transExp(e.head);
     else {
        ExpTy head = transExp(e.head);
        ExpTy tail = transExpSeq(e.tail);
        return new ExpTy(translate.seqExp(head.exp, tail.exp), tail.ty);
     }
 }

 ExpTy transExp(Absyn.SeqExp e) {
     return transExpSeq(e.list);
 }

 ExpTy transExp(Absyn.AssignExp e) {
   ExpTy v = transVar(e.var);
   return new ExpTy(translate.assignExp(v.exp, coerce(transExp(e.exp), v.ty,e.pos)),
	        VOID);
 }

 ExpTy transExp(Absyn.IfExp e) {
     Exp test = checkInt(transExp(e.test), e.test.pos);
     ExpTy thenclause = transExp(e.thenclause);
     if (e.elseclause == null)
        return new ExpTy(translate.ifThen(test, coerce(thenclause, VOID, 
					           e.thenclause.pos)),
		      VOID);
     else return new ExpTy(translate.ifThenElse(test,
				thenclause.exp,
				coerce(transExp(e.elseclause),
				       thenclause.ty, 
				       e.elseclause.pos)),
			thenclause.ty);
 }

 ExpTy transExp(Absyn.NilExp e) {
   return new ExpTy(translate.intExp(0), new Types.NIL());
 }

 ExpTy transExp(Absyn.ForExp e) {
   Absyn.Var i = new Absyn.SimpleVar(e.var.pos,e.var.name);
   Symbol intsym = Symbol.symbol("int");
   Symbol hisym = Symbol.symbol(".hi.");
   Absyn.Var hivar = new Absyn.SimpleVar(e.hi.pos,hisym);
   Absyn.VarDec hidec = new Absyn.VarDec(e.hi.pos,hisym,new Absyn.NameTy(e.hi.pos,intsym),e.hi);
   Absyn.Exp incr = new Absyn.AssignExp(e.hi.pos,i,new Absyn.OpExp(e.hi.pos,new Absyn.VarExp(e.hi.pos,i),
						  Absyn.OpExp.PLUS,
						  new Absyn.IntExp(e.hi.pos,1)));
   Absyn.Exp wh = new Absyn.WhileExp(e.pos,new Absyn.OpExp(e.pos,new Absyn.VarExp(e.var.pos,i),
					     Absyn.OpExp.LE,
					     new Absyn.VarExp(e.hi.pos,hivar)),
				 new Absyn.SeqExp(e.pos,new Absyn.ExpList(e.body,
					       new Absyn.ExpList(incr,null))));
   hidec.escape=false;
   return transExp(new Absyn.LetExp(e.pos,new Absyn.DecList(e.var,
					new Absyn.DecList(hidec,null)),
				wh));
				

 }

 ExpTy transExp(Absyn.WhileExp e) {
     BreakScope done = new BreakScope();
     Semant sem = new Semant(env,translate,level,done);
     return new ExpTy(translate.whileExp(checkInt(sem.transExp(e.test), e.pos),
				     coerce(sem.transExp(e.body),VOID,e.body.pos),
				     done),
		  VOID);
 }

 ExpTy transExp(Absyn.BreakExp e) {
  if (breakScope == null) {
     error(e.pos, "break not within loop");
     return bogusExp();
  } else {
     return new ExpTy(translate.breakExp(breakScope), VOID);
  }
 }

 ExpTy transExp(Absyn.LetExp e) {
     env.venv.beginScope();
     env.tenv.beginScope();
     Exp decs = translate.nop();
     for (Absyn.DecList p=e.decs; p!=null; p=p.tail)
	decs = translate.seqExp(decs, transDec(p.head));
     ExpTy et = transExp(e.body);
     env.venv.endScope();
     env.tenv.endScope();
     return new ExpTy(translate.seqExp(decs,et.exp), et.ty);
 }

 ExpTy transExp(Absyn.ArrayExp e) {
    Types.Type x = tylook(e.typ,e.pos).actual();
    if (x instanceof Types.ARRAY) {
       Types.ARRAY arr = (Types.ARRAY)x;
       return new ExpTy(translate.arrayExp(checkInt(transExp(e.size), e.pos),
				coerce(transExp(e.init), arr.element, e.pos)),
	     arr);
    } else {
       error(e.pos, "array-ID not an array type"); 
       return bogusExp();
    }
 }
    
 ExpTy transExp(Absyn.Exp e) {
   if (e instanceof Absyn.VarExp) return transExp((Absyn.VarExp)e);
   if (e instanceof Absyn.IntExp) return transExp((Absyn.IntExp)e);
   if (e instanceof Absyn.StringExp) return transExp((Absyn.StringExp)e);
   if (e instanceof Absyn.CallExp) return transExp((Absyn.CallExp)e);
   if (e instanceof Absyn.OpExp) return transExp((Absyn.OpExp)e);
   if (e instanceof Absyn.RecordExp) return transExp((Absyn.RecordExp)e);
   if (e instanceof Absyn.SeqExp) return transExp((Absyn.SeqExp)e);
   if (e instanceof Absyn.AssignExp) return transExp((Absyn.AssignExp)e);
   if (e instanceof Absyn.IfExp) return transExp((Absyn.IfExp)e);
   if (e instanceof Absyn.NilExp) return transExp((Absyn.NilExp)e);
   if (e instanceof Absyn.ForExp) return transExp((Absyn.ForExp)e);
   if (e instanceof Absyn.WhileExp) return transExp((Absyn.WhileExp)e);
   if (e instanceof Absyn.BreakExp) return transExp((Absyn.BreakExp)e);
   if (e instanceof Absyn.LetExp) return transExp((Absyn.LetExp)e);
   if (e instanceof Absyn.ArrayExp) return transExp((Absyn.ArrayExp)e);
   throw new Error("transExp: " + e.getClass().toString());
 }

  Exp transDec(Absyn.VarDec d) {
      ExpTy e = transExp(d.init);
      Translate.Access access = level.allocLocal(d.escape);

      if (d.typ != null) {
       Types.Type t = transTy(d.typ);
       e = new ExpTy(coerce(e,t,d.init.pos), t);
      }
      env.venv.put(d.name, new VarEntry(access,e.ty));
      return translate.assignExp(translate.varExp(
			   translate.simpleVar(access,level,d.name)),
			   e.exp);
  }

  Exp transDec(Absyn.FunctionDec d) {
        OneFunc funcs = new OneFunc(this, d);
	for(OneFunc f=funcs; f!=null; f=f.next)
           env.venv.put(f.dec.name, 
		  new FunEntry(level, f.label, f.formals, f.result));
	for(OneFunc f=funcs; f!=null; f=f.next)
          f.doBody();
        checkDups(funcs);
       return translate.nop();
    }

  Exp transDec(Absyn.TypeDec d) {
       OneType types = new OneType(d);
       for (OneType t=types; t!=null; t=t.next)
          env.tenv.put(t.dec.name, t.namety);
       for (OneType t=types; t!=null; t=t.next)
          t.namety.bind(transTy(t.dec.ty));
       /* verify that every loop goes through Record or Array */
       for (OneType t=types; t!=null; t=t.next)
          if (t.namety.isLoop())
             error(t.dec.pos, "illegal circular definition");
       checkDups(d);
       return translate.nop();
  }

 Exp transDec(Absyn.Dec d) {
   if (d instanceof Absyn.VarDec)      return transDec((Absyn.VarDec)d);
   if (d instanceof Absyn.FunctionDec) return transDec((Absyn.FunctionDec)d);
   if (d instanceof Absyn.TypeDec)     return transDec((Absyn.TypeDec)d);
   throw new Error("transDec");   
 }

 public Translate.Frag transProg(Absyn.Exp exp) {
     Level lev = new Level(level, Symbol.symbol("toymain"), null);
     Semant sem = new Semant(env,translate,lev,null);
     ExpTy et = sem.transExp(exp);
     Exp body = et.exp;
     translate.procEntryExit(lev,body);
     {Translate.Frag frags = translate.getResult();
      if (env.errorMsg.anyErrors) return null;
      else return frags;
     }
 }

}

class OneFunc {
    Semant semant;
    Absyn.FunctionDec dec;
    Label label;
    OneFunc next;
    Types.RECORD formals;
    Types.Type result;

    OneFunc(Semant sem, Absyn.FunctionDec d) {
      semant=sem;
      dec=d;
      label = new Label(d.name);
      formals = semant.transTypeFields(dec.params);
      result = dec.result != null
                            ? semant.transTy(dec.result)
                            : semant.VOID;
      
      if (d.next != null) next = new OneFunc(semant, d.next);
    }

    Util.BoolList escapeList(Absyn.FieldList params) {
         if (params==null) return null;
         else return new Util.BoolList(params.escape,
				       escapeList(params.tail));
    }
 
    void doBody() {
      semant.env.venv.beginScope();
      Level lev = new Level(semant.level, dec.name, 
					escapeList(dec.params));
      Translate.AccessList p; 
      Absyn.FieldList r;
      for(p=lev.formals, r=dec.params; p!=null; p=p.tail, r=r.tail)
        semant.env.venv.put(r.name, 
	         new VarEntry(p.head, semant.tylook(r.typ,r.pos)));
      Semant sem = new Semant(semant.env,semant.translate,lev,
				(BreakScope)null);
      semant.translate.procEntryExit(lev,
		sem.coerce(sem.transExp(dec.body), result, dec.pos));
      semant.env.venv.endScope();
     }
}


class OneType {
   Types.NAME namety;
   Absyn.TypeDec dec;
   OneType next;

   OneType(Absyn.TypeDec d) {
     namety = new Types.NAME(d.name);
     dec=d;
     next = d.next!=null ? new OneType(d.next) : null;
   }
}
